; Copyright (c) 2019-2021 Intel Corporation
;
; Licensed under the Apache License, Version 2.0 (the "License");
; you may not use this file except in compliance with the License.
; You may obtain a copy of the License at
;
;     http://www.apache.org/licenses/LICENSE-2.0
;
; Unless required by applicable law or agreed to in writing, software
; distributed under the License is distributed on an "AS IS" BASIS,
; WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
; See the License for the specific language governing permissions and
; limitations under the License.

EXPORTS

global
__TBB_internal_initialize_system_topology
__TBB_internal_apply_affinity
__TBB_internal_restore_affinity
__TBB_internal_allocate_binding_handler
__TBB_internal_deallocate_binding_handler
__TBB_internal_get_default_concurrency
__TBB_internal_destroy_system_topology
